//----------------------------------------------------------------------------------
// aviutl  DXVA2 C^[[XۑTCYtB^
// Copyright (c) 2007  N099
//----------------------------------------------------------------------------------
#include <windows.h>

#include <math.h>
#include <float.h>
#include <malloc.h>
#include <mmintrin.h>
#include <xmmintrin.h>
#include <new>

#include <initguid.h>
#include <d3d9.h>
#include <dxva2api.h>
#include <dxerr.h>

#include "resource.h"
#include "filter.h"

#include "cnvthrd.h"
#include "dxva2_proc.h"

//---------------------------------------------------------------------
// internal function prototypes
//---------------------------------------------------------------------
static bool isSSE2usable(void);
static CNV_TYPE toCnvType(D3DFORMAT fmt, bool sub, bool toPYC);
static int toBpp(CNV_TYPE type);

//---------------------------------------------------------------------
// const
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
DXVAFilter::DXVAFilter(bool sse2) :
	m_pD3D(NULL), m_pd3dDevice(NULL), m_hDXVA2DLL(0),
	m_pDXVAVProcService(NULL), m_pDXVAVProcDevice(NULL),
	m_pConv(NULL)
{
	m_mode = m_WidthIn = m_HeightIn = m_WidthOut = m_HeightOut = m_wMax = 0;

	m_ConvWorkers = 1;

	m_ascend = false;

	ZeroMemory(m_pSrfSrc, sizeof(m_pSrfSrc));
	ZeroMemory(m_pSrfDst, sizeof(m_pSrfDst));
	ZeroMemory(m_pSrfDstTmp, sizeof(m_pSrfDstTmp));

	//T[tFXtH[}bg
	m_SrcFormat = m_DstFormat = D3DFMT_YUY2;
	m_SrcCnvType = PYC_YUY2;
	m_DstCnvType = YUY2_PYC;

	//SSE2eXg
	m_sse2 = sse2 ? isSSE2usable() : sse2;
}

//fXgN^
DXVAFilter::~DXVAFilter()
{
	//\[X̊J
	ReleaseResources();
	DestroyDXVA2();

	//dxva2.dll
	if(m_hDXVA2DLL){
		FreeLibrary(m_hDXVA2DLL);
	}

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);
}

//IuWFNg̏
bool DXVAFilter::Init(const FILTER *fp)
{
	//D3D̏
	if(FAILED(InitD3D(fp->hwnd))){
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}

	//DXVA2̏
	SAFE_RELEASE(m_pDXVAVProcService);
	if(FAILED(InitDXVA2Service())){
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}

	return true;
}

//e탊\[X̊m
bool DXVAFilter::AllocResources(const DXVAFilterParam *para)
{
	HRESULT hr = S_OK;

	//ƈ`FbN
	if(!m_pDXVAVProcService || !para) return false;

	//sNZTCỸZbgi32*32 -- 4096x4096ɐj
	if(para->HeightIn<32 || para->HeightIn>4096 || para->WidthIn<32 || para->WidthIn>4096
		|| para->HeightOut<32 || para->HeightOut>4096 || para->WidthOut<32 || para->WidthOut>4096
		|| para->wMax < para->WidthIn || para->wMax < para->WidthOut)
	{
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("Ή̉摜TCYł"));
		return false;
	}else{
		m_WidthIn   = para->WidthIn;
		m_HeightIn  = para->HeightIn;
		m_WidthOut  = para->WidthOut;
		m_HeightOut = para->HeightOut;
		m_wMax      = para->wMax;
	}

	//̂ق
	m_mode      = para->mode;
	m_SrcFormat = para->SrcFormat;
	m_DstFormat = para->DstFormat;
	m_ascend    = para->SrcAscend;
	
	m_SrcCnvType = toCnvType(m_SrcFormat, para->SrcFormatSub, false);
	m_DstCnvType = toCnvType(m_DstFormat, para->DstFormatSub, true);
	if(m_SrcCnvType == EXIT_CNV || m_DstCnvType == EXIT_CNV){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("Ή̃tH[}bgł"));
		return false;
	}

	//ꉞJĂ
	ReleaseResources();

	//
	// rfIvZbT̍쐬
	//
	hr = CreateDXVA2Processor();
	if(FAILED(hr)){
		ReleaseResources();
		return false;
	}

	//T[tFX̍쐬
	hr = CreateSurface();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFX쐬Ɏs܂"));
		ReleaseResources();
		return false;
	}

	//Œp[^̏
	InitProcParams();

	//PIXEL_YCϊXbh̍쐬
	m_pConv = new(std::nothrow) CConv[2];
	if(NULL == m_pConv){
		ReleaseResources();
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("ŝ悤ł"));
		return false;
	}else{
		//ϊXbh̋N
		for(int i=0; i<2; ++i){
			if(!m_pConv[i].Create(m_ConvWorkers, m_sse2)){
				ReleaseResources();
				ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("ϊXbh̏Ɏs܂"));
				return false;
			}
		}
	}

	return true;
}

//e탊\[X̊J
bool DXVAFilter::ReleaseResources(void)
{	
	//\[X̊J
	SAFE_DELETE_ARRAY(m_pConv);

	DestroySurface();

	SAFE_RELEASE(m_pDXVAVProcDevice);

	return true;
}

//f[^̏s
bool DXVAFilter::ProcImage(PIXEL_YC *dst, const PIXEL_YC *src, int frame, bool tff, int out)
{
	HRESULT hr;

	if(!m_pd3dDevice || !m_pDXVAVProcDevice){
		return false;
	}

	// D3D9 device ̏ԂeXg
	hr = m_pd3dDevice->TestCooperativeLevel();
	switch (hr)
	{
	case D3D_OK:
		break;
	case D3DERR_DEVICELOST:
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DERR_DEVICELOST"));
		return false;
	case D3DERR_DEVICENOTRESET:
		hr = ResetDevice();
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("DeviceReset error."));
			return false;
		}
		break;
	default :
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("TestCooperativeLevel error."));
		return false;
	}

	//TvύXȂ
	if(src != NULL){
		if(FAILED(AppendFrame(src, frame, tff))){
			return false;
		}
	}

	// Sample
	int n = m_ascend ? 0 : (N_SAMP - (m_Caps.NumBackwardRefSamples + 1));
	DXVA2_VideoSample *vs = m_BltSample + n;
	n = N_SAMP - n; //Tv

	//Tv̂
	if(dst == NULL){
		//nvidian͓ŃTvۗ?
		ProcBlt0(vs, n, frame, tff);
		return true;
	}

	//ProcessBltŏ
	hr = E_FAIL;
	if(out == 0){
		hr = ProcBlt(dst, vs, n, frame, tff);
	}else if(out == 1){
		hr = ProcBlt1(dst, vs, n, frame, tff);
	}else if(out == 2){
		hr = ProcBlt2(dst, vs, n, frame, tff);
	}

	return FAILED(hr) ? false : true;
}

//D3DC^[tF[XOɓn
bool DXVAFilter::GetD3DInterface(IDirect3D9 **ppD3D, IDirect3DDevice9 **ppd3dDevice)
{
	if(!m_pD3D || !m_pd3dDevice || !ppD3D || !ppd3dDevice){
		return false;
	}else{
		//AddRefsŃRs[
		m_pD3D->AddRef();
		m_pd3dDevice->AddRef();
		*ppD3D = m_pD3D;
		*ppd3dDevice = m_pd3dDevice; 
	}

	return true;
}

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//D3D9̏
HRESULT DXVAFilter::InitD3D(HWND hwnd)
{
	HRESULT hr;

	//Ŝ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	//init D3D    
	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(NULL == m_pD3D) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("Direct3DCreate9()s܂"));
		return E_OUTOFMEMORY;
	}

	//create device
	D3DPRESENT_PARAMETERS d3dpp; 

	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pD3D->CreateDevice(
		D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_NOWINDOWCHANGES, 
		&d3dpp, &m_pd3dDevice );
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DfoCX̍쐬Ɏs܂"));
	}

	return hr;
}

//DXVA2̏
HRESULT DXVAFilter::InitDXVA2Service(void)
{
	HRESULT hr;
	HRESULT (WINAPI * DXVA2CreateVS)(IDirect3DDevice9*, REFIID, void**) = NULL;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//DXVA2płȂꍇ̂ŖIɃ[h
	if(!m_hDXVA2DLL){
		m_hDXVA2DLL = LoadLibrary(TEXT("dxva2.dll"));
	}

	if(m_hDXVA2DLL){
		DXVA2CreateVS = reinterpret_cast<HRESULT (WINAPI *)(IDirect3DDevice9*, REFIID, void**)>
			(GetProcAddress(m_hDXVA2DLL, "DXVA2CreateVideoService"));
	}

	if(NULL == DXVA2CreateVS){
		if(m_hDXVA2DLL){
			FreeLibrary(m_hDXVA2DLL);
			m_hDXVA2DLL = NULL;
		}
		hr = E_NOINTERFACE;
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("DXVA2pł܂"));
		return hr;
	}

	//DXVA2 Video Processor Service ̍쐬
	SAFE_RELEASE(m_pDXVAVProcService);
	hr = DXVA2CreateVS(m_pd3dDevice, IID_IDirectXVideoProcessorService,
		reinterpret_cast<VOID**>(&m_pDXVAVProcService));
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("Video Processor Service pł܂"));
		return hr;
	}
	
	return hr;
}

//rfIvZbT̍쐬
HRESULT DXVAFilter::CreateDXVA2Processor(void)
{	
	//|C^̃`FbN
	if(!m_pDXVAVProcService) return D3DERR_INVALIDCALL;

	// VideoDesc̍쐬
	ZeroMemory(&m_VideoDesc, sizeof(m_VideoDesc));
	m_VideoDesc.SampleWidth                         = m_WidthIn;
	m_VideoDesc.SampleHeight                        = m_HeightIn;
	m_VideoDesc.SampleFormat.VideoChromaSubsampling = DXVA2_VideoChromaSubsampling_MPEG2;
	m_VideoDesc.SampleFormat.NominalRange           = DXVA2_NominalRange_16_235;
	m_VideoDesc.SampleFormat.VideoTransferMatrix    = DXVA2_VideoTransferMatrix_BT601;
	m_VideoDesc.SampleFormat.VideoLighting          = DXVA2_VideoLighting_dim;
	m_VideoDesc.SampleFormat.VideoPrimaries         = DXVA2_VideoPrimaries_BT709;
	m_VideoDesc.SampleFormat.VideoTransferFunction  = DXVA2_VideoTransFunc_709;
	m_VideoDesc.SampleFormat.SampleFormat           = DXVA2_SampleFieldInterleavedEvenFirst;
	m_VideoDesc.Format                              = m_SrcFormat;
	m_VideoDesc.InputSampleFreq.Numerator           = 25;
	m_VideoDesc.InputSampleFreq.Denominator         = 1;
	m_VideoDesc.OutputFrameFreq.Numerator           = 50;
	m_VideoDesc.OutputFrameFreq.Denominator         = 1;

	// ΉvZbT GUID ₢킹
	UINT count;
	GUID* guids = NULL;
	HRESULT hr;

	hr = m_pDXVAVProcService->GetVideoProcessorDeviceGuids(&m_VideoDesc, &count, &guids);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("p\ȃrfIvZbT܂"));
		return hr;
	}

	//DXVA2 device ̍쐬
	SAFE_RELEASE(m_pDXVAVProcDevice);
	for(UINT i = (m_mode>=count ? (count-1) : m_mode); i < count; i++){
		hr = CreateDXVA2Device(guids[i]);
		if(S_OK == hr){
			break;
		}
	}
	CoTaskMemFree(guids);

	if(!m_pDXVAVProcDevice){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("rfIvZbT̍쐬Ɏs܂"));
		return hr;
	}

	return hr;
}

//rfIvZbTfoCX̍쐬
HRESULT DXVAFilter::CreateDXVA2Device(const GUID& guid)
{
	HRESULT hr;

	// T|[gtH[}bg`FbN
	UINT i, count;
	D3DFORMAT* formats = NULL;

	hr = m_pDXVAVProcService->GetVideoProcessorRenderTargets(guid, &m_VideoDesc, &count, &formats);
	if(FAILED(hr)){
		return hr;
	}

	for(i=0; i<count; i++){
		if(formats[i] == m_DstFormat){
			break;
		}
	}
	CoTaskMemFree(formats);

	if(i >= count){
		return E_FAIL;
	}

	//
	// vZbT@\eXg
	//
	hr = m_pDXVAVProcService->GetVideoProcessorCaps(guid, &m_VideoDesc, m_DstFormat, &m_Caps);
	if(FAILED(hr)){
		return hr;
	}
#if 0
	// \tgEFAfoCX͗pȂ
	if(m_Caps.DeviceCaps & DXVA2_VPDev_SoftwareDevice){
		return E_FAIL;
	}
#endif
	// T|[g@\̃`FbN
	UINT op = DXVA2_VideoProcess_StretchX | DXVA2_VideoProcess_StretchY;
	if(isProcRGB()){
		op |= DXVA2_VideoProcess_YUV2RGB | DXVA2_VideoProcess_YUV2RGBExtended;
	}
	if((m_Caps.VideoProcessorOperations & op) != op){
		return E_FAIL;
	}

	//tB^p[^擾ĕۑ
	hr = GetFilterRange(guid);
	if(FAILED(hr)){
		return hr;
	}

	//
	// video processor device ̍쐬
	//
	return m_pDXVAVProcService->CreateVideoProcessor(
		guid, &m_VideoDesc, m_DstFormat, 0, &m_pDXVAVProcDevice);
}

//DXVA2\[X̊J
void DXVAFilter::DestroyDXVA2(void)
{
	SAFE_RELEASE(m_pDXVAVProcDevice);
	SAFE_RELEASE(m_pDXVAVProcService);
}

//T[tFX̊m
HRESULT DXVAFilter::CreateSurface(void)
{
	HRESULT hr = S_OK;
	int i;

	//܂͊J
	DestroySurface();

	if(!m_pDXVAVProcService){
		return D3DERR_INVALIDCALL;
	}

	//Src
	for(i=0; i<ARRAYSIZE(m_pSrfSrc); ++i){
		hr = m_pDXVAVProcService->CreateSurface(m_WidthIn, m_HeightIn, 0, m_SrcFormat,
			m_Caps.InputPool, 0, DXVA2_VideoProcessorRenderTarget, &m_pSrfSrc[i], NULL);
		if(FAILED(hr)){
			return hr;
		}
	}

	//Dst
	for(i=0; i<ARRAYSIZE(m_pSrfDst); ++i){
		hr = m_pd3dDevice->CreateRenderTarget(m_WidthOut, m_HeightOut, m_DstFormat,
			D3DMULTISAMPLE_NONE, 0, FALSE, &m_pSrfDst[i], NULL);
		if(FAILED(hr)){
			return hr;
		}
	}

	//Dst temp
	for(i=0; i<ARRAYSIZE(m_pSrfDstTmp); ++i){
		hr = m_pd3dDevice->CreateOffscreenPlainSurface(
			m_WidthOut, m_HeightOut, m_DstFormat, D3DPOOL_SYSTEMMEM, &m_pSrfDstTmp[i], NULL);
		if(FAILED(hr)){
			return hr;
		}
	}

	return hr;
}

//ProcAmp,Filter̃ftHgl擾
HRESULT DXVAFilter::GetFilterRange(const GUID& guid)
{
	UINT i;
	DXVA2_ValueRange range;
	HRESULT hr = S_OK;

	if(!m_pDXVAVProcService){
		return D3DERR_INVALIDCALL;
	}

	// ProcAmp range
	ZeroMemory(m_ProcAmpRanges, sizeof(m_ProcAmpRanges));
	ZeroMemory(m_ProcAmpValues, sizeof(m_ProcAmpValues));
	for(i = 0; i < ARRAYSIZE(m_ProcAmpRanges); i++){
		if(m_Caps.ProcAmpControlCaps & (1 << i)){
			hr = m_pDXVAVProcService->GetProcAmpRange(guid, &m_VideoDesc, m_DstFormat, 1 << i, &range);
			if(FAILED(hr)){
				return hr;
			}

			//WƃftHglZbg
			m_ProcAmpRanges[i] = range;
			m_ProcAmpValues[i] = range.DefaultValue;
		}
	}

	// Noise Filter range
	ZeroMemory(m_NFilterRanges, sizeof(m_NFilterRanges));
	ZeroMemory(m_NFilterValues, sizeof(m_NFilterValues));
	if(m_Caps.VideoProcessorOperations & DXVA2_VideoProcess_NoiseFilter){
		for(i = 0; i < ARRAYSIZE(m_NFilterRanges); i++){
			hr = m_pDXVAVProcService->GetFilterPropertyRange(guid, &m_VideoDesc, m_DstFormat,
				DXVA2_NoiseFilterLumaLevel + i,	&range);
			if(FAILED(hr))
			{		
				return hr;
			}

			//WƃftHglZbg
			m_NFilterRanges[i] = range;
			m_NFilterValues[i] = range.DefaultValue;
		}
	}

	// Detail Filter range
	ZeroMemory(m_DFilterRanges, sizeof(m_DFilterRanges));
	ZeroMemory(m_DFilterValues, sizeof(m_DFilterValues));
	if(m_Caps.VideoProcessorOperations & DXVA2_VideoProcess_DetailFilter){
		for(i = 0; i < ARRAYSIZE(m_DFilterRanges); i++){
			hr = m_pDXVAVProcService->GetFilterPropertyRange(guid, &m_VideoDesc, m_DstFormat,
				DXVA2_DetailFilterLumaLevel + i, &range);
			if(FAILED(hr)){
				return hr;
			}

			//WƃftHglZbg
			m_DFilterRanges[i] = range;
			m_DFilterValues[i] = range.DefaultValue;
		}
	}

	return hr;
}

//t[ԂŕωȂp[^Q̏
void DXVAFilter::InitProcParams(void)
{
	DXVA2_AYUVSample16 bg = { 0, 0, 0, 0xFFFF };

	//
	// init VPBlt parameters
	//
	ZeroMemory(&m_BltParam, sizeof(m_BltParam));
	SetRect(&m_BltParam.TargetRect, 0, 0, m_WidthOut, m_HeightOut);

	// DXVA2_VideoProcess_Constriction
	m_BltParam.ConstrictionSize.cx = m_WidthOut;
	m_BltParam.ConstrictionSize.cy = m_HeightOut;

	m_BltParam.BackgroundColor = bg;

	UINT subsamp = (isProcRGB() ? DXVA2_VideoChromaSubsampling_Unknown : DXVA2_VideoChromaSubsampling_Cosited);
	UINT nrange  = (isProcRGB() ? DXVA2_NominalRange_0_255 : DXVA2_NominalRange_16_235);

	// DXVA2_VideoProcess_YUV2RGBExtended
	m_BltParam.DestFormat.VideoChromaSubsampling = subsamp;
	m_BltParam.DestFormat.NominalRange           = nrange;
	m_BltParam.DestFormat.VideoTransferMatrix    = DXVA2_VideoTransferMatrix_BT601;
	m_BltParam.DestFormat.VideoLighting          = DXVA2_VideoLighting_dim;
	m_BltParam.DestFormat.VideoPrimaries         = DXVA2_VideoPrimaries_BT709;
	m_BltParam.DestFormat.VideoTransferFunction  = DXVA2_VideoTransFunc_709;

	m_BltParam.DestFormat.SampleFormat = DXVA2_SampleProgressiveFrame;
	m_BltParam.Alpha = DXVA2_Fixed32OpaqueAlpha();

	// DXVA2_ProcAmp
	m_BltParam.ProcAmpValues.Brightness = m_ProcAmpValues[0];
	m_BltParam.ProcAmpValues.Contrast   = m_ProcAmpValues[1];
	m_BltParam.ProcAmpValues.Hue        = m_ProcAmpValues[2];
	m_BltParam.ProcAmpValues.Saturation = m_ProcAmpValues[3];

	// DXVA2_VideoProcess_NoiseFilter
	m_BltParam.NoiseFilterLuma.Level       = m_NFilterValues[0];
	m_BltParam.NoiseFilterLuma.Threshold   = m_NFilterValues[1];
	m_BltParam.NoiseFilterLuma.Radius      = m_NFilterValues[2];
	m_BltParam.NoiseFilterChroma.Level     = m_NFilterValues[3];
	m_BltParam.NoiseFilterChroma.Threshold = m_NFilterValues[4];
	m_BltParam.NoiseFilterChroma.Radius    = m_NFilterValues[5];

	// DXVA2_VideoProcess_DetailFilter
	m_BltParam.DetailFilterLuma.Level       = m_DFilterValues[0];
	m_BltParam.DetailFilterLuma.Threshold   = m_DFilterValues[1];
	m_BltParam.DetailFilterLuma.Radius      = m_DFilterValues[2];
	m_BltParam.DetailFilterChroma.Level     = m_DFilterValues[3];
	m_BltParam.DetailFilterChroma.Threshold = m_DFilterValues[4];
	m_BltParam.DetailFilterChroma.Radius    = m_DFilterValues[5];

	//
	// Initialize main stream video sample.
	//
	DXVA2_VideoSample samp;
	ZeroMemory(&samp, sizeof(samp));

	// DXVA2_VideoProcess_YUV2RGBExtended (input is YUV Format)
	samp.SampleFormat.VideoChromaSubsampling = DXVA2_VideoChromaSubsampling_MPEG2;
	samp.SampleFormat.NominalRange           = DXVA2_NominalRange_16_235;
	samp.SampleFormat.VideoTransferMatrix    = DXVA2_VideoTransferMatrix_BT601;
	samp.SampleFormat.VideoLighting          = DXVA2_VideoLighting_dim;
	samp.SampleFormat.VideoPrimaries         = DXVA2_VideoPrimaries_BT709;
	samp.SampleFormat.VideoTransferFunction  = DXVA2_VideoTransFunc_709;

	// DXVA2_VideoProcess_SubRects
	SetRect(&samp.SrcRect, 0, 0, m_WidthIn, m_HeightIn);
	
	// DXVA2_VideoProcess_StretchX, Y
	SetRect(&samp.DstRect, 0, 0, m_WidthOut, m_HeightOut);

	// DXVA2_VideoProcess_PlanarAlpha
	samp.PlanarAlpha = DXVA2_Fixed32OpaqueAlpha();

	for(int i=0; i<ARRAYSIZE(m_BltSample); ++i){
		m_BltSample[i] = samp;
		m_BltSample[i].SrcSurface = m_pSrfSrc[i];
	}
}

//D3DT[tFX֓]
HRESULT DXVAFilter::PutPixelData(IDirect3DSurface9 *dst, const PIXEL_YC *src, CConv &cv)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;
	CNV_PARAM cp;

	hr = dst->LockRect(&lRect, NULL, D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	cp.type = m_SrcCnvType;
	cp.top = false;
	cp.inPtr = (void*)src;
	cp.outPtr = lRect.pBits;
	cp.h = m_HeightIn;
	cp.w = m_WidthIn;
	cp.inStride = m_wMax;
	cp.outStride = lRect.Pitch/toBpp(cp.type);

	if(!cv.Request(&cp)){
		dst->UnlockRect();
		return E_FAIL;
	}

	return hr;
}

//D3DT[tFX]
HRESULT DXVAFilter::GetPixelData(PIXEL_YC *dst, IDirect3DSurface9 *src, CConv &cv, bool top)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;
	CNV_PARAM cp;

	hr = src->LockRect(&lRect, NULL, D3DLOCK_NOSYSLOCK|D3DLOCK_READONLY);
	if(hr != D3D_OK) return hr;

	cp.type = m_DstCnvType;
	cp.top = top;
	cp.inPtr = top ? lRect.pBits : ((BYTE*)lRect.pBits + lRect.Pitch);
	cp.outPtr = top ? dst : (dst + m_wMax);
	cp.h = m_HeightOut/2;
	cp.w = m_WidthOut;
	cp.inStride = (2 * lRect.Pitch)/toBpp(m_DstCnvType);
	cp.outStride = m_wMax * 2;

	if(!cv.Request(&cp)){
		src->UnlockRect();
		return E_FAIL;
	}

	return hr;
}

//D3DT[tFX]iŚj
HRESULT DXVAFilter::GetPixelData(PIXEL_YC *dst, IDirect3DSurface9 *src, CConv &cv)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;
	CNV_PARAM cp;

	hr = src->LockRect(&lRect, NULL, D3DLOCK_NOSYSLOCK|D3DLOCK_READONLY);
	if(hr != D3D_OK) return hr;

	cp.type = m_DstCnvType;
	cp.top = true;
	cp.inPtr = lRect.pBits;
	cp.outPtr = dst;
	cp.h = m_HeightOut;
	cp.w = m_WidthOut;
	cp.inStride = lRect.Pitch/toBpp(m_DstCnvType);
	cp.outStride = m_wMax;

	if(!cv.Request(&cp)){
		src->UnlockRect();
		return E_FAIL;
	}

	return hr;
}

//t[Tv
HRESULT DXVAFilter::AppendFrame(const PIXEL_YC *src, int frame, bool tff)
{
	//TvXCh
	SwapSample();

	DXVA2_VideoSample *pvs = SampleToAppend();

	HRESULT hr = PutPixelData(pvs->SrcSurface, src, m_pConv[0]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("PutPixelData error."));
		return hr;
	}

	// Set video sample
	pvs->Start = toRefTime(frame, 25);
	pvs->End   = toRefTime(frame + 1, 25);
	pvs->SampleData = tff ? DXVA2_SampleData_TFF : 0;
	pvs->SampleFormat.SampleFormat =
		tff ? DXVA2_SampleFieldInterleavedEvenFirst : DXVA2_SampleFieldInterleavedOddFirst;

	m_pConv[0].Wait(); //#0-0
	pvs->SrcSurface->UnlockRect();

	return hr;
}

//ProcessBltsiC^[[X𕜌j
HRESULT DXVAFilter::ProcBlt(PIXEL_YC *dst, DXVA2_VideoSample *vs, int n, int frame, bool tff)
{
	HRESULT hr;

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame, 25);

	// 1st
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[0], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.1"));
	}

	hr = m_pd3dDevice->GetRenderTargetData(m_pSrfDst[0], m_pSrfDstTmp[0]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetRenderTargetData error.1"));
	}

	hr = GetPixelData(dst, m_pSrfDstTmp[0], m_pConv[0], tff); //#1-0
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetPixelData error.1"));
	}

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame * 2 + 1, 50);

	// 2nd
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[1], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.2"));
	}

	hr = m_pd3dDevice->GetRenderTargetData(m_pSrfDst[1], m_pSrfDstTmp[1]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetRenderTargetData error.2"));
	}

	hr = GetPixelData(dst, m_pSrfDstTmp[1], m_pConv[1], !tff); //#1-1
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetPixelData error.2"));
	}

	m_pConv[0].Wait(); //#1-0
	m_pSrfDstTmp[0]->UnlockRect();

	m_pConv[1].Wait(); //#1-1
	m_pSrfDstTmp[1]->UnlockRect();

	return hr;
}

//ProcessBltsiʂ͎擾Ȃj
HRESULT DXVAFilter::ProcBlt0(DXVA2_VideoSample *vs, int n, int frame, bool tff)
{
	HRESULT hr;

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame, 25);

	// 1st
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[0], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.1"));
	}

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame * 2 + 1, 50);

	// 2ndiʂ͎擾Ȃj
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[1], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.2"));
	}

	return hr;
}

//ProcessBltsistB[ĥ݁j
HRESULT DXVAFilter::ProcBlt1(PIXEL_YC *dst, DXVA2_VideoSample *vs, int n, int frame, bool tff)
{
	HRESULT hr;

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame, 25);

	// 1st
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[0], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.1"));
	}

	hr = m_pd3dDevice->GetRenderTargetData(m_pSrfDst[0], m_pSrfDstTmp[0]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetRenderTargetData error.1"));
	}

	hr = GetPixelData(dst, m_pSrfDstTmp[0], m_pConv[0]); //#1-0
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetPixelData error.1"));
	}

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame * 2 + 1, 50);

	// 2ndiʂ͎擾Ȃj
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[1], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.2"));
	}

	m_pConv[0].Wait(); //#1-0
	m_pSrfDstTmp[0]->UnlockRect();

	return hr;
}

//ProcessBltsĩtB[ĥ݁j
HRESULT DXVAFilter::ProcBlt2(PIXEL_YC *dst, DXVA2_VideoSample *vs, int n, int frame, bool tff)
{
	HRESULT hr;

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame, 25);

	// 1stiʂ͎擾Ȃj
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[0], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.1"));
	}

	// VPBlt param
	m_BltParam.TargetFrame = toRefTime(frame * 2 + 1, 50);

	// 2nd
	hr = m_pDXVAVProcDevice->VideoProcessBlt(m_pSrfDst[1], &m_BltParam, vs, n, NULL);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VideoProcessBlt error.2"));
	}

	hr = m_pd3dDevice->GetRenderTargetData(m_pSrfDst[1], m_pSrfDstTmp[1]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetRenderTargetData error.2"));
	}

	hr = GetPixelData(dst, m_pSrfDstTmp[1], m_pConv[1]); //#1-1
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GetPixelData error.2"));
	}

	m_pConv[1].Wait(); //#1-1
	m_pSrfDstTmp[1]->UnlockRect();

	return hr;
}

//DEVICE_LOST̕A
HRESULT DXVAFilter::ResetDevice(void)
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return D3DERR_NOTAVAILABLE;
	}

	DestroySurface();
	DestroyDXVA2();

	//foCXZbg
	D3DPRESENT_PARAMETERS dpp;
	hr = m_pd3dDevice->Reset(&dpp);
	if(FAILED(hr)){
		return hr;
	}

	hr = InitDXVA2Service();
	if(FAILED(hr)){
		return hr;
	}

	// rfIvZbT̍쐬
	hr = CreateDXVA2Processor();
	if(FAILED(hr)){
		return hr;
	}

	//T[tFX̍쐬
	hr = CreateSurface();
	if(FAILED(hr)){
		return hr;
	}

	return hr;
}

//G[̕\
HRESULT DXVAFilter::ShowMessageBox(const char *file, DWORD line, HRESULT hr, const TCHAR *msg)
{
#ifdef _DEBUG
	return DXTrace(file, line, hr, msg, TRUE);
#else
	MessageBox(NULL, msg, NULL, MB_ICONERROR|MB_SYSTEMMODAL);
	return S_OK;
#endif
}

//̓T[tFX̓ւ
void DXVAFilter::SwapSample(void)
{
	DXVA2_VideoSample vtmp;

	if(m_ascend){
		vtmp = m_BltSample[ARRAYSIZE(m_BltSample) - 1];
		MoveMemory(m_BltSample+1, m_BltSample, sizeof(m_BltSample) - sizeof(m_BltSample[0]));
		m_BltSample[0] = vtmp;
	}else{
		vtmp = m_BltSample[0];
		MoveMemory(m_BltSample, m_BltSample+1, sizeof(m_BltSample) - sizeof(m_BltSample[0]));
		m_BltSample[ARRAYSIZE(m_BltSample) - 1] = vtmp;
	}
}

//T[tFX̊J
void DXVAFilter::DestroySurface(void)
{
	int i;
	for(i=0; i<ARRAYSIZE(m_pSrfSrc); ++i){
		SAFE_RELEASE(m_pSrfSrc[i]);
	}

	for(i=0; i<ARRAYSIZE(m_pSrfDst); ++i){
		SAFE_RELEASE(m_pSrfDst[i]);
	}

	for(i=0; i<ARRAYSIZE(m_pSrfDstTmp); ++i){
		SAFE_RELEASE(m_pSrfDstTmp[i]);
	}
}

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------
//SSE2ɑΉĂ邩`FbN
static bool isSSE2usable(void)
{
	return IsProcessorFeaturePresent(PF_XMMI64_INSTRUCTIONS_AVAILABLE) ? true : false;
}

//tH[}bgϊ^Cv쐬
CNV_TYPE toCnvType(D3DFORMAT fmt, bool sub, bool toPYC)
{
	if(D3DFMT_YUY2 == fmt){
		if(toPYC){
			return sub ? YUY2S_PYC : YUY2_PYC;
		}else{
			return sub ? PYC_YUY2S : PYC_YUY2;
		}
	}else if(D3DFMT_X8R8G8B8 == fmt || D3DFMT_A8R8G8B8 == fmt){
		if(toPYC){
			return sub ? RGB32S_PYC : RGB32_PYC;
		}
	}else if(fmt == D3DFMT_NV12){
		if(!toPYC){
			return sub ? PYC_NV12S : PYC_NV12;
		}
	}
	
	//Ή
	return EXIT_CNV;
}

//ϊΏۃtH[}bg̃sNZ̃oCg
static int toBpp(CNV_TYPE type)
{
	const int tbl[EXIT_CNV] = { 2, 2, 1, 1, 2, 2, 4, 4 };
	if(type >= PYC_YUY2 && type < EXIT_CNV){
		return tbl[type];
	}

	return 4;
}
