/* convert.c
 * color converter YUY2 <-> PIXEL_YC
 */
#include <emmintrin.h>

#include "convert.h"

#define CLIP(x) (unsigned char)((x) >= 255 ? 255 : ((x) < 0 ? 0 : x))

/* ACg킹p */
typedef union _u128 {
    short s[8];
    __m128i x;
} u128;

typedef union _u64 {
    short s[4];
    __m64 m;
} u64;

/*
 *  PIXEL_YC -> YUY2 left origin (w/o LPF)
 */
int pyc_to_yuy2(unsigned char *dst, const unsigned short *src,
                unsigned int h, unsigned int w,
                unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int y,c;
    const PIXEL_YC *src_p = (const PIXEL_YC *)src;

    w = (w + 1)/2;
    is = (is - w * 2);
    os = (os - w * 2) * 2;

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            y = (src_p->y * 219 + 383 + (16 << 12)) >> 12;
            *dst++ = CLIP(y);
            c = ((src_p->cb + 2048) * 7 + 66 + (16 << 7)) >> 7;
            *dst++ = CLIP(c);
            c = ((src_p->cr + 2048) * 7 + 66 + (16 << 7)) >> 7;
            src_p++;
            y = (src_p->y * 219 + 383 + (16 << 12)) >> 12;
            *dst++ = CLIP(y);
            *dst++ = CLIP(c);
            src_p++;
        }
        src_p += is;
        dst += os;
    }

    return 0;
}

/*
 *  PIXEL_YC -> YUY2(full scale) left origin (w/o LPF)
 */
int pyc_to_yuy2s(unsigned char *dst, const unsigned short *src,
                 unsigned int h, unsigned int w,
                 unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int y,c;
    const PIXEL_YC *src_p = (const PIXEL_YC *)src;

    w = (w + 1)/2;
    is = (is - w * 2);
    os = (os - w * 2) * 2;

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            y = src_p->y >> 4;
            *dst++ = CLIP(y);
            c = (src_p->cb + 2048) >> 4;
            *dst++ = CLIP(c);
            c = (src_p->cr + 2048) >> 4;
            src_p++;
            y = src_p->y >> 4;
            *dst++ = CLIP(y);
            *dst++ = CLIP(c);
            src_p++;
        }
        src_p += is;
        dst += os;
    }

    return 0;
}

/*
 *  PIXEL_YC -> YUY2 left origin (w/o LPF) SSE2
 */
int pyc_to_yuy2_SSE2(unsigned char *dst, const unsigned short *src,
                     unsigned int h, unsigned int w,
                     unsigned int is, unsigned int os)
{
    unsigned int i,j;
    const PIXEL_YC *src_p;
    __m64 *mdst;
    __m128i xi0,xi1,xi2,xf,xe,xm;
    static const u128 uof = { 383, 14402, 383, 14402, 383, 14402, 383, 14402 };
    static const u128 uoe = {  16,  16,  16,    16,  16,  16,  16,    16 };
    static const u128 uml = { 219,   1, 224,    32, 219,   1, 224,    32 };

    /* 4pixel𓯎ɏ */
    w = ((w + 3)/4) - 1;
    xf = uof.x;
    xe = uoe.x;
    xm = uml.x;
    xi2 = _mm_setzero_si128();

    /* 1[vǂݍ݂s */
    src_p = (const PIXEL_YC *)src;

    xi1 = _mm_loadl_epi64((__m128i*)src_p);
    src_p += 2;
    xi0 = _mm_loadl_epi64((__m128i*)src_p);
    src_p += 2;

    xi1 = _mm_shufflelo_epi16(xi1, 0xB4);
    xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
    xi1 = _mm_unpacklo_epi64(xi1, xi0);
    
    for(i=0; i<h; ++i){
        mdst  = (__m64*)(dst + 2 * i * os);
        for(j=0; j<w; ++j){
            /* 񃋁[v̂߂̓ǂݍ */
            xi2 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;
            xi0 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;

            xi2 = _mm_shufflelo_epi16(xi2, 0xB4);
            xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
            xi2 = _mm_unpacklo_epi64(xi2, xi0);

            /* WC^[[u */
            xi0 = _mm_unpackhi_epi16(xi1, xf);
            xi1 = _mm_unpacklo_epi16(xi1, xf);
            
            /* ZƃItZbgZ */
            xi0 = _mm_madd_epi16(xi0, xm);
            xi1 = _mm_madd_epi16(xi1, xm);
            xi0 = _mm_srai_epi32(xi0, 12);
            xi1 = _mm_srai_epi32(xi1, 12);
            
            /* 4pixelpbN */
            xi1 = _mm_packs_epi32(xi1, xi0);
            xi1 = _mm_adds_epi16(xi1, xe);
            xi1 = _mm_packus_epi16(xi1, xi1);

            /* 64bitXgA */
            _mm_stream_pi(mdst++, _mm_movepi64_pi64(xi1));

            xi1 = xi2;
        }
        /* E[͓ʂɈ */
        if(i < h-1){
            src_p = (const PIXEL_YC *)src + (i+1) * is;
            
            xi2 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;
            xi0 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;

            xi2 = _mm_shufflelo_epi16(xi2, 0xB4);
            xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
            xi2 = _mm_unpacklo_epi64(xi2, xi0);
        }
        
        /* WC^[[u */
        xi0 = _mm_unpackhi_epi16(xi1, xf);
        xi1 = _mm_unpacklo_epi16(xi1, xf);
        
        /* ZƃItZbgZ */
        xi0 = _mm_madd_epi16(xi0, xm);
        xi1 = _mm_madd_epi16(xi1, xm);
        xi0 = _mm_srai_epi32(xi0, 12);
        xi1 = _mm_srai_epi32(xi1, 12);
        
        /* 4pixelpbN */
        xi1 = _mm_packs_epi32(xi1, xi0);
        xi1 = _mm_adds_epi16(xi1, xe);
        xi1 = _mm_packus_epi16(xi1, xi1);
        
        /* 64bitXgA */
        _mm_stream_pi(mdst++, _mm_movepi64_pi64(xi1));
        
        xi1 = xi2;
    }
    
    _mm_sfence();
    _mm_empty();
    
    return 0;
}

/*
 *  PIXEL_YC -> YUY2(full) left origin (w/o LPF) SSE2
 */
int pyc_to_yuy2s_SSE2(unsigned char *dst, const unsigned short *src,
                      unsigned int h, unsigned int w,
                      unsigned int is, unsigned int os)
{
    unsigned int i,j;
    const PIXEL_YC *src_p;
    __m64 *mdst;
    __m128i xi0,xi1,xi2,xs;
    static const u128 uofs = { 0, 2048, 0, 2048, 0, 2048, 0, 2048 };

    /* 4pixel𓯎ɏ */
    w = ((w + 3)/4) - 1;
    xs = uofs.x;
    xi2 = _mm_setzero_si128();

    /* 1[vǂݍ݂s */
    src_p = (const PIXEL_YC *)src;

    xi1 = _mm_loadl_epi64((__m128i*)src_p);
    src_p += 2;
    xi0 = _mm_loadl_epi64((__m128i*)src_p);
    src_p += 2;

    xi1 = _mm_shufflelo_epi16(xi1, 0xB4);
    xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
    xi1 = _mm_unpacklo_epi64(xi1, xi0);
    
    for(i=0; i<h; ++i){
        mdst  = (__m64*)(dst + 2 * i * os);
        for(j=0; j<w; ++j){
            /* 񃋁[v̂߂̓ǂݍ */
            xi2 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;
            xi0 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;

            xi2 = _mm_shufflelo_epi16(xi2, 0xB4);
            xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
            xi2 = _mm_unpacklo_epi64(xi2, xi0);

            xi0 = xi1;
            
            /* ItZbgZăpbN */
            xi0 = _mm_adds_epi16(xi0, xs);
            xi0 = _mm_srai_epi16(xi0, 4);
            xi0 = _mm_packus_epi16(xi0, xi0);

            xi1 = xi2;
            
            /* 64bitXgA */
            _mm_stream_pi(mdst++, _mm_movepi64_pi64(xi0));
        }
        /* E[͓ʂɈ */
        if(i < h-1){
            src_p = (const PIXEL_YC *)src + (i+1) * is;
            
            xi2 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;
            xi0 = _mm_loadl_epi64((__m128i*)src_p);
            src_p += 2;

            xi2 = _mm_shufflelo_epi16(xi2, 0xB4);
            xi0 = _mm_shufflelo_epi16(xi0, 0xB4);
            xi2 = _mm_unpacklo_epi64(xi2, xi0);
        }
        xi0 = xi1;
            
        /* ItZbgZăpbN */
        xi0 = _mm_adds_epi16(xi0, xs);
        xi0 = _mm_srai_epi16(xi0, 4);
        xi0 = _mm_packus_epi16(xi0, xi0);
        
        xi1 = xi2;
        
        /* 64bitXgA */
        _mm_stream_pi(mdst++, _mm_movepi64_pi64(xi0));
    }
    
    _mm_sfence();
    _mm_empty();
    
    return 0;
}

/*
 *  YUY2 -> PIXEL_YC left origin iאڃsNZԁj
 */
int yuy2_to_pyc(unsigned short *dst, const unsigned char *src,
                unsigned int h, unsigned int w,
                unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int cb0,cb1,cr0,cr1;
    PIXEL_YC *dst_p = (PIXEL_YC *)dst;

    w = (w + 1)/2;
    is = (is - w * 2) * 2;
    os = (os - w * 2);

    for(i=0; i<h; ++i){
        cb0 = cb1 = src[1];
        cr0 = cr1 = src[3];
        for(j=1; j<w; ++j){
            dst_p->y  = (short)((*src++ * 1197 - (299 << 6)) >> 6);
            src++; //cb
            dst_p->cb = (short)(((cb1 - 128) * 4681 + 164) >> 8);
            dst_p->cr = (short)(((cr1 - 128) * 4681 + 164) >> 8);
            dst_p++;
            
            dst_p->y  = (short)((*src++ * 1197 - (299 << 6)) >> 6);
            src++; //cr
            
            cb0 = cb1;
            cr0 = cr1;
            cb1 = src[1];
            cr1 = src[3];
            cb0 = (cb1 + cb0 + 1)/2;
            cr0 = (cr1 + cr0 + 1)/2;
            
            dst_p->cb = (short)(((cb0 - 128) * 4681 + 164) >> 8);
            dst_p->cr = (short)(((cr0 - 128) * 4681 + 164) >> 8);
            dst_p++;
        }
        /* right */
        cb0 = ((cb1 - 128) * 4681 + 164) >> 8;
        cr0 = ((cr1 - 128) * 4681 + 164) >> 8;
        dst_p->y  = (short)((*src++ * 1197 - (299 << 6)) >> 6);
        src++; //cb
        dst_p->cb = (short)cb0;
        dst_p->cr = (short)cr0;
        dst_p++;
        
        dst_p->y  = (short)((*src++ * 1197 - (299 << 6)) >> 6);
        src++; //cr
        dst_p->cb = (short)cb0;
        dst_p->cr = (short)cr0;
        dst_p++;
        
        src += is;
        dst_p += os;
    }

    return 0;
}

/*
 *  YUY2 -> PIXEL_YC left origin iאڃsNZԁj SSE2
 */
int yuy2_to_pyc_SSE2(unsigned short *dst, const unsigned char *src,
                     unsigned int h, unsigned int w,
                     unsigned int is, unsigned int os)
{
    unsigned int i,j;
    __m64 *msrc,*mdst;
    __m128i xy0,xc0,xc1,xs0,xm0,xm1,xt0,xt1;
    static const u128 uoa = { 1, 1, 1, 1, 1, 1, 1, 1 };
    static const u128 uma = { 1197, -19136, 1197, -19136, 1197, -19136, 1197, -19136 };
    static const u128 umb = { 4681, 164, 4681, 164, 4681, 164, 4681, 164 };

    /* 4pixel𓯎ɏ悤Ƃ */
    if(w & 0x3U){
        /* 4̔{ȊO͒ʏ֐ŏ */
        return yuy2_to_pyc(dst, src, h, w ,is, os);
    }

    w /= 4;
    xs0 = uoa.x;
    xm0 = uma.x;
    xm1 = umb.x;

    msrc = (__m64*)src;
    xt0 = _mm_loadl_epi64((__m128i*)msrc++);
    xt1 = _mm_loadl_epi64((__m128i*)msrc);
    
    for(i=0; i<h; ++i){
        mdst = (__m64*)(dst + 3 * i * os);
        for(j=1; j<w; j++){
            xt1 = _mm_loadl_epi64((__m128i*)msrc++);

            /* Pxo */
            xy0 = _mm_slli_epi16(xt0, 8);
            xy0 = _mm_srli_epi16(xy0, 8);

            /* Fo */
            xc1 = _mm_srli_si128(xt0, 4); /* 0-0 0 YC2 */
            xc0 = _mm_srli_epi16(xt0, 8); /* 0-0 C2 C0 */
            xc1 = _mm_unpacklo_epi32(xc1, xt1); /* YC6 0 YC4 YC2 */
            xc1 = _mm_srli_epi16(xc1, 8); /* C6 0 C4 C2 */
                       
            /* Px */
            xy0 = _mm_unpacklo_epi16(xy0, xs0);
            xy0 = _mm_madd_epi16(xy0, xm0);
            xy0 = _mm_srai_epi32(xy0, 6); /* Y3 Y2 Y1 Y0 */
            xy0 = _mm_packs_epi32(xy0, xy0);

            /* ςĕ */
            xt0 = _mm_slli_epi16(xs0, 7); /* 128 */
            xc1 = _mm_avg_epu16(xc1, xc0); /* C6/2 0 C3 C1 */
            xc0 = _mm_sub_epi16(xc0, xt0); /* X-X C2 C0 */
            xc1 = _mm_sub_epi16(xc1, xt0); /* X-X C3 C1 */
 
            /* F */
            xc0 = _mm_unpacklo_epi16(xc0, xs0);
            xc1 = _mm_unpacklo_epi16(xc1, xs0);
            xc0 = _mm_madd_epi16(xc0, xm1);
            xc1 = _mm_madd_epi16(xc1, xm1);
            xc0 = _mm_srai_epi32(xc0, 8);
            xc1 = _mm_srai_epi32(xc1, 8);
            xc0 = _mm_packs_epi32(xc0, xc1);

            /* xy0 : Y_3 Y_2 Y_1 Y_0 Y_3 Y_2 Y_1 Y_0 */
            /* xc0 : Cr3 Cb3 Cr1 Cb1 Cr2 Cb2 Cr0 Cb0 */
            
            /* C^[[u */
            xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
            xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */
            
            xt0 = xt1;

            /* ёւ */
            xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
            xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
            xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
            
            /* store */
            _mm_storel_epi64((__m128i*)mdst++, xy0);
            _mm_storeu_si128((__m128i*)mdst, xc0);
            mdst += 2;
        }
        /* E[ */
        if(i < h-1){
            msrc = (__m64*)(src + 2 * (i + 1) * is);
            xt1 = _mm_loadl_epi64((__m128i*)msrc++);
        }
        
        /* Pxo */
        xy0 = _mm_slli_epi16(xt0, 8);
        xy0 = _mm_srli_epi16(xy0, 8);
        
        /* Fo */
        xc0 = _mm_srli_epi16(xt0, 8); /* 0-0 C2 C0 */
        xc1 = _mm_shufflelo_epi16(xc0, 0xEE); /* 0-0 C2 C2 */
        
        /* Px */
        xy0 = _mm_unpacklo_epi16(xy0, xs0);
        xy0 = _mm_madd_epi16(xy0, xm0);
        xy0 = _mm_srai_epi32(xy0, 6); /* Y3 Y2 Y1 Y0 */
        xy0 = _mm_packs_epi32(xy0, xy0);
        
        /* ςĕ */
        xt0 = _mm_slli_epi16(xs0, 7); /* 128 */
        xc1 = _mm_avg_epu16(xc1, xc0); /* C6/2 0 C3 C1 */
        xc0 = _mm_sub_epi16(xc0, xt0); /* X-X C2 C0 */
        xc1 = _mm_sub_epi16(xc1, xt0); /* X-X C3 C1 */
        
        /* F */
        xc0 = _mm_unpacklo_epi16(xc0, xs0);
        xc1 = _mm_unpacklo_epi16(xc1, xs0);
        xc0 = _mm_madd_epi16(xc0, xm1);
        xc1 = _mm_madd_epi16(xc1, xm1);
        xc0 = _mm_srai_epi32(xc0, 8);
        xc1 = _mm_srai_epi32(xc1, 8);
        xc0 = _mm_packs_epi32(xc0, xc1);
        
        /* xy0 : Y_3 Y_2 Y_1 Y_0 Y_3 Y_2 Y_1 Y_0 */
        /* xc0 : Cr3 Cb3 Cr1 Cb1 Cr2 Cb2 Cr0 Cb0 */
            
        /* C^[[u */
        xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
        xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */
        
        xt0 = xt1;
        
        /* ёւ */
        xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
        xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
        xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
        
        /* store */
        _mm_storel_epi64((__m128i*)mdst++, xy0);
        _mm_storeu_si128((__m128i*)mdst, xc0);
        mdst += 2;
    }
    
    return 0;
}

/*
 *  YUY2(full scale) -> PIXEL_YC left origin iאڃsNZԁj
 */
int yuy2s_to_pyc(unsigned short *dst, const unsigned char *src,
                 unsigned int h, unsigned int w,
                 unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int cb0,cb1,cr0,cr1;
    PIXEL_YC *dst_p = (PIXEL_YC *)dst;

    w = (w + 1)/2;
    is = (is - w * 2) * 2;
    os = (os - w * 2);

    for(i=0; i<h; ++i){
        cb0 = cb1 = src[1];
        cr0 = cr1 = src[3];
        for(j=1; j<w; ++j){
            dst_p->y  = (short)(*src++ << 4);
            src++; //cb
            dst_p->cb = (short)((cb1 - 128) << 4);
            dst_p->cr = (short)((cr1 - 128) << 4);
            dst_p++;
            
            dst_p->y  = (short)(*src++ << 4);
            src++; //cr
            
            cb0 = cb1;
            cr0 = cr1;
            cb1 = src[1];
            cr1 = src[3];
            cb0 = (cb1 + cb0 + 1)/2;
            cr0 = (cr1 + cr0 + 1)/2;
            
            dst_p->cb = (short)((cb0 - 128) << 4);
            dst_p->cr = (short)((cr0 - 128) << 4);
            dst_p++;
        }
        /* right */
        cb0 = (cb1 - 128) << 4;
        cr0 = (cr1 - 128) << 4;
        dst_p->y  = (short)(*src++ << 4);
        src++; //cb
        dst_p->cb = (short)cb0;
        dst_p->cr = (short)cr0;
        dst_p++;
        
        dst_p->y  = (short)(*src++ << 4);
        src++; //cr
        dst_p->cb = (short)cb0;
        dst_p->cr = (short)cr0;
        dst_p++;
        
        src += is;
        dst_p += os;
    }

    return 0;
}

/*
 *  YUY2(full) -> PIXEL_YC left origin iאڃsNZԁj SSE2
 *  !! 8̔{̂ݑΉ !!
 */
int yuy2s_to_pyc_SSE2(unsigned short *dst, const unsigned char *src,
                      unsigned int h, unsigned int w,
                      unsigned int is, unsigned int os)
{
    unsigned int i,j;
    __m64 *msrc,*mdst;
    __m128i xy0,xc0,xc1,xs0,xs1,xt0,xt1;
    static const u128 uoa = { 255, 255, 255, 255, 0, 0, 0, 0 };
    static const u128 uob = { 128, 128, 128, 128, 128, 128, 128, 128 };

    /* 4pixel𓯎ɏ悤Ƃ */
    if(w & 0x3U){
        /* 4̔{ȊO͒ʏ֐ŏ */
        return yuy2s_to_pyc(dst, src, h, w ,is, os);
    }

    w /= 4;
    xs0 = uoa.x;
    xs1 = uob.x;
  
    msrc = (__m64*)src;
    xt0 = _mm_loadl_epi64((__m128i*)msrc++);
    xt1 = _mm_loadl_epi64((__m128i*)msrc);
    
    for(i=0; i<h; ++i){
        mdst = (__m64*)(dst + 3 * i * os);
        for(j=1; j<w; j++){
            xt1 = _mm_loadl_epi64((__m128i*)msrc++);

            /* Pxo */
            xy0 = _mm_and_si128(xt0, xs0);

            /* Fo */
            xc1 = _mm_srli_si128(xt0, 4); /* 0-0 0 YC2 */
            xc0 = _mm_srli_epi16(xt0, 8); /* 0-0 C2 C0 */
            xc1 = _mm_unpacklo_epi32(xc1, xt1); /* YC6 0 YC4 YC2 */
            xc1 = _mm_srli_epi16(xc1, 8); /* C6 0 C4 C2 */
                       
            /* Px */
            xy0 = _mm_slli_epi16(xy0, 4); /* Y3 Y2 Y1 Y0 */
            xy0 = _mm_unpacklo_epi64(xy0, xy0);

            /* ςĕ */
            xc1 = _mm_avg_epu16(xc1, xc0); /* C6/2 0 C3 C1 */
 
            /* F */
            xc0 = _mm_unpacklo_epi64(xc0, xc1); /* C3 C1 C2 C0 */
            xc0 = _mm_sub_epi16(xc0, xs1);
            xc0 = _mm_slli_epi16(xc0, 4);

            /* xy0 : Y_3 Y_2 Y_1 Y_0 Y_3 Y_2 Y_1 Y_0 */
            /* xc0 : Cr3 Cb3 Cr1 Cb1 Cr2 Cb2 Cr0 Cb0 */
            
            /* C^[[u */
            xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
            xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */
            
            xt0 = xt1;

            /* ёւ */
            xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
            xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
            xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
            
            /* store */
            _mm_storel_epi64((__m128i*)mdst++, xy0);
            _mm_storeu_si128((__m128i*)mdst, xc0);
            mdst += 2;
        }
        /* E[ */
        if(i < h-1){
            msrc = (__m64*)(src + 2 * (i + 1) * is);
            xt1 = _mm_loadl_epi64((__m128i*)msrc++);
        }
        
        /* Pxo */
        xy0 = _mm_and_si128(xt0, xs0);

        /* Fo */
        xc1 = _mm_srli_si128(xt0, 4); /* 0-0 0 YC2 */
        xc0 = _mm_srli_epi16(xt0, 8); /* 0-0 C2 C0 */
        xc1 = _mm_unpacklo_epi32(xc1, xt1); /* YC6 0 YC4 YC2 */
        xc1 = _mm_srli_epi16(xc1, 8); /* C6 0 C4 C2 */
        
        /* Px */
        xy0 = _mm_slli_epi16(xy0, 4); /* Y3 Y2 Y1 Y0 */
        xy0 = _mm_unpacklo_epi64(xy0, xy0);

        /* ςĕ */
        xc1 = _mm_avg_epu16(xc1, xc0); /* C6/2 0 C3 C1 */
 
        /* F */
        xc0 = _mm_unpacklo_epi64(xc0, xc1); /* C3 C1 C2 C0 */
        xc0 = _mm_sub_epi16(xc0, xs1);
        xc0 = _mm_slli_epi16(xc0, 4);
        
        /* xy0 : Y_3 Y_2 Y_1 Y_0 Y_3 Y_2 Y_1 Y_0 */
        /* xc0 : Cr3 Cb3 Cr1 Cb1 Cr2 Cb2 Cr0 Cb0 */
            
        /* C^[[u */
        xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
        xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */
        
        xt0 = xt1;
        
        /* ёւ */
        xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
        xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
        xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
        
        /* store */
        _mm_storel_epi64((__m128i*)mdst++, xy0);
        _mm_storeu_si128((__m128i*)mdst, xc0);
        mdst += 2;
    }

    return 0;
}

/*
 *  RGB32 -> PIXEL_YC(BT.601)
 */
int rgb32_to_pyc(unsigned short *dst, const unsigned char *src,
                 unsigned int h, unsigned int w,
                 unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int b,g,r;
    PIXEL_YC *dst_p = (PIXEL_YC *)dst;

    is = (is - w) * 4;
    os = os - w;

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            b = *src++;
            g = *src++;
            r = *src++;
            src++; /* a */
            dst_p->y  = (short)((b *  467 + g * 2404 + r * 1225) >> 8);
            dst_p->cb = (short)((b * 2048 - g * 1357 - r *  691) >> 8);
            dst_p->cr = (short)((b * -333 - g * 1715 + r * 2048) >> 8);
            dst_p++;
        }
        src += is;
        dst_p += os;
    }

    return 0;
}

/*
 *  RGB32 -> PIXEL_YC(BT.709)
 */
int rgb32s_to_pyc(unsigned short *dst, const unsigned char *src,
                  unsigned int h, unsigned int w,
                  unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int b,g,r;
    PIXEL_YC *dst_p = (PIXEL_YC *)dst;

    is = (is - w) * 4;
    os = os - w;

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            b = *src++;
            g = *src++;
            r = *src++;
            src++; /* a */
            dst_p->y  = (short)((b *  295 + g * 2930 + r *  870) >> 8);
            dst_p->cb = (short)((b * 2048 - g * 1579 - r *  469) >> 8);
            dst_p->cr = (short)((b * -188 - g * 1866 + r * 2055) >> 8);
            dst_p++;
        }
        src += is;
        dst_p += os;
    }

    return 0;
}

/*
 *  RGB32 -> PIXEL_YC SSE2(BT.601)
 */
int rgb32_to_pyc_SSE2(unsigned short *dst, const unsigned char *src,
                      unsigned int h, unsigned int w,
                      unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int b,g,r;
    PIXEL_YC *dst_p;
    const int *src_i;
    const unsigned char *src_c;
    __m128i xi0,xi1,xt0,xy0,xc0,xc1,xm0,xm1,xm2;
    static const u128 um0 = { 2404, 0, 2404, 0, 467, 1225, 467, 1225 };
    static const u128 um1 = { -1357, 0, -1357, 0, 2048, -691, 2048, -691 };
    static const u128 um2 = { -1715, 0, -1715, 0,  -333, 2048, -333, 2048 };

	/* 4pixel𓯎ɏ悤Ƃ */
    if(w & 0x3U){
        /* 4̔{ȊO͒ʏ֐ŏ */
        return rgb32_to_pyc(dst, src, h, w ,is, os);
    }

    xm0 = um0.x;
    xm1 = um1.x;
    xm2 = um2.x;
    
    for(i=0; i<h; ++i){
        src_i = (const int*)src + is * i;
        dst_p = (PIXEL_YC*)dst + os * i;

        /* 4pixel𓯎ɏ */
        for(j=w ; j>4; j-=4){
            xi0 = _mm_load_si128((__m128i*)src_i);
            src_i += 4;
            
            xi1 = _mm_slli_epi16(xi0, 8);
            xi0 = _mm_srli_epi16(xi0, 8); /* A,G */
            xi1 = _mm_srli_epi16(xi1, 8); /* R,B */

            /* Y */
            xt0 = _mm_unpacklo_epi64(xm0, xm0);
            xy0 = _mm_unpackhi_epi64(xm0, xm0);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xy0 = _mm_madd_epi16(xy0, xi1);
            
            xy0 = _mm_add_epi32(xy0, xt0);
            xy0 = _mm_srai_epi32(xy0, 8);
            
            xy0 = _mm_packs_epi32(xy0, xy0);
            
            /* Cb */
            xt0 = _mm_unpacklo_epi64(xm1, xm1);
            xc0 = _mm_unpackhi_epi64(xm1, xm1);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xc0 = _mm_madd_epi16(xc0, xi1);
            
            xc0 = _mm_add_epi32(xc0, xt0);
            xc0 = _mm_srai_epi32(xc0, 8);
            
            /* Cr */
            xt0 = _mm_unpacklo_epi64(xm2, xm2);
            xc1 = _mm_unpackhi_epi64(xm2, xm2);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xc1 = _mm_madd_epi16(xc1, xi1);
            
            xc1 = _mm_add_epi32(xc1, xt0);
            xc1 = _mm_srai_epi32(xc1, 8);

            xt0 = _mm_unpackhi_epi32(xc0, xc1); /* Cr3 Cb3 Cr2 Cb2 */
            xc0 = _mm_unpacklo_epi32(xc0, xc1); /* Cr1 Cb1 Cr0 Cb0 */
            
            xc1 = _mm_unpackhi_epi64(xc0, xt0); /* Cr3 Cb3 Cr2 Cb2 */
            xc0 = _mm_unpacklo_epi64(xc0, xt0); /* Cr1 Cb1 Cr0 Cb0 */
            
            xc0 = _mm_packs_epi32(xc0, xc1); /* Cr3 Cb3 Cr2 Cb2 Cr1 Cb1 Cr0 Cb0 */

            /* C^[[u */
            xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
            xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */

            /* ёւ */
            xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
            xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
            xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
            
            /* store */
            _mm_storel_epi64((__m128i*)dst_p, xy0);
            _mm_storeu_si128((__m128i*)((__m64*)dst_p + 1), xc0);
            dst_p += 4;
        }
        
        /* c */
        for( ; j>0; --j){
            src_c = (const unsigned char*)src_i;
            b = src_c[0];
            g = src_c[1];
            r = src_c[2];
            src_i++;
            dst_p->y  = (short)((b *  467 + g * 2404 + r * 1225) >> 8);
            dst_p->cb = (short)((b * 2048 - g * 1357 - r *  691) >> 8);
            dst_p->cr = (short)((b * -333 - g * 1715 + r * 2048) >> 8);
            dst_p++;
        }
    }

    _mm_empty();
    
    return 0;
}

/*
 *  RGB32 -> PIXEL_YC SSE2(BT.709)
 */
int rgb32s_to_pyc_SSE2(unsigned short *dst, const unsigned char *src,
                       unsigned int h, unsigned int w,
                       unsigned int is, unsigned int os)
{
    unsigned int i,j;
    int b,g,r;
    PIXEL_YC *dst_p;
    const int *src_i;
    const unsigned char *src_c;
    __m128i xi0,xi1,xt0,xy0,xc0,xc1,xm0,xm1,xm2;
    static const u128 um0 = { 2930, 0, 2930, 0, 295, 870, 295, 870 };
    static const u128 um1 = { -1579, 0, -1579, 0, 2048, -469, 2048, -469 };
    static const u128 um2 = { -1866, 0, -1866, 0, -188, 2055, -188, 2055 };

    /* 4pixel𓯎ɏ悤Ƃ */
    if(w & 0x3U){
        /* 4̔{ȊO͒ʏ֐ŏ */
        return rgb32s_to_pyc(dst, src, h, w ,is, os);
    }

    xm0 = um0.x;
    xm1 = um1.x;
    xm2 = um2.x;
    
    for(i=0; i<h; ++i){
        src_i = (const int*)src + is * i;
        dst_p = (PIXEL_YC*)dst + os * i;

        /* 4pixel𓯎ɏ */
        for(j=w ; j>4; j-=4){
            xi0 = _mm_load_si128((__m128i*)src_i);
            src_i += 4;
            
            xi1 = _mm_slli_epi16(xi0, 8);
            xi0 = _mm_srli_epi16(xi0, 8); /* A,G */
            xi1 = _mm_srli_epi16(xi1, 8); /* R,B */

            /* Y */
            xt0 = _mm_unpacklo_epi64(xm0, xm0);
            xy0 = _mm_unpackhi_epi64(xm0, xm0);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xy0 = _mm_madd_epi16(xy0, xi1);
            
            xy0 = _mm_add_epi32(xy0, xt0);
            xy0 = _mm_srai_epi32(xy0, 8);
            
            xy0 = _mm_packs_epi32(xy0, xy0);
            
            /* Cb */
            xt0 = _mm_unpacklo_epi64(xm1, xm1);
            xc0 = _mm_unpackhi_epi64(xm1, xm1);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xc0 = _mm_madd_epi16(xc0, xi1);
            
            xc0 = _mm_add_epi32(xc0, xt0);
            xc0 = _mm_srai_epi32(xc0, 8);
            
            /* Cr */
            xt0 = _mm_unpacklo_epi64(xm2, xm2);
            xc1 = _mm_unpackhi_epi64(xm2, xm2);
            
            xt0 = _mm_madd_epi16(xt0, xi0);
            xc1 = _mm_madd_epi16(xc1, xi1);
            
            xc1 = _mm_add_epi32(xc1, xt0);
            xc1 = _mm_srai_epi32(xc1, 8);

            xt0 = _mm_unpackhi_epi32(xc0, xc1); /* Cr3 Cb3 Cr2 Cb2 */
            xc0 = _mm_unpacklo_epi32(xc0, xc1); /* Cr1 Cb1 Cr0 Cb0 */
            
            xc1 = _mm_unpackhi_epi64(xc0, xt0); /* Cr3 Cb3 Cr2 Cb2 */
            xc0 = _mm_unpacklo_epi64(xc0, xt0); /* Cr1 Cb1 Cr0 Cb0 */
            
            xc0 = _mm_packs_epi32(xc0, xc1); /* Cr3 Cb3 Cr2 Cb2 Cr1 Cb1 Cr0 Cb0 */

            /* C^[[u */
            xy0 = _mm_unpacklo_epi16(xy0, xc0); /* Cr2 Y_3 Cb2 Y_2 Cr0 Y_1 Cb0 Y_0 */
            xc0 = _mm_unpackhi_epi16(xc0, xy0); /* Cr2 Cr3 Y_3 Cb3 Cb2 Cr1 Y_2 Cb1 */

            /* ёւ */
            xc0 = _mm_shufflehi_epi16(xc0, 0x87); /* Cr3 Cb3 Y_3 Cr2 */
            xy0 = _mm_shufflelo_epi16(xy0, 0xB4); /* Y_1 Cr0 Cb0 Y_0 */
            xc0 = _mm_shufflelo_epi16(xc0, 0xD8); /* Cb2 Y_2 Cr1 Cb1 */
            
            /* store */
            _mm_storel_epi64((__m128i*)dst_p, xy0);
            _mm_storeu_si128((__m128i*)((__m64*)dst_p + 1), xc0);
            dst_p += 4;
        }
        
        /* c */
        for( ; j>0; --j){
            src_c = (const unsigned char*)src_i;
            b = src_c[0];
            g = src_c[1];
            r = src_c[2];
            src_i++;
            dst_p->y  = (short)((b *  295 + g * 2930 + r *  870) >> 8);
            dst_p->cb = (short)((b * 2048 - g * 1579 - r *  469) >> 8);
            dst_p->cr = (short)((b * -188 - g * 1866 + r * 2055) >> 8);
            dst_p++;
        }
    }

    return 0;
}

int pyc_to_nv12(unsigned char *dst, const unsigned short *src,
                unsigned int h, unsigned int w,
                unsigned int is, unsigned int os)
{
    unsigned int i,j,k,l = is * 2;
    int y,c;
    const PIXEL_YC *src_p = (const PIXEL_YC *)src;

    w = (w + 1)/2;
    is = (is - w * 2);
    os = (os - w * 2);

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            y = (src_p->y * 219 + 383 + (16 << 12)) >> 12;
            *dst++ = CLIP(y);
            src_p++;
            y = (src_p->y * 219 + 383 + (16 << 12)) >> 12;
            *dst++ = CLIP(y);
            src_p++;
        }
        src_p += is;
        dst += os;
    }

	src_p = (const PIXEL_YC *)src;
	h /= 2;
	k = h & 1;
	h -= k;
    for(i=0; i<h; ++i){
		if(i&1){ //bottom
			for(j=0; j<w; ++j){
				//c = (src_p->cb + 3 * (src_p + l)->cb)/4;
				c = (src_p + l)->cb;
				c = ((c + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				//c = (src_p->cr + 3 * (src_p + l)->cr)/4;
				c = (src_p + l)->cr;
				c = ((c + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is + l;
			dst += os;
		}else{ //top
			for(j=0; j<w; ++j){
				//c = (3 * src_p->cb + (src_p + l)->cb)/4;
				c = src_p->cb;
				c = ((c + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				//c = (3 * src_p->cr + (src_p + l)->cr)/4;
				c = src_p->cr;
				c = ((c + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is;
			dst += os;
		}
    }
	if(k){
		for(i=0; i<2; ++i){
			for(j=0; j<w; ++j){
				c = ((src_p->cb + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				c = ((src_p->cr + 2048) * 7 + 66 + (16 << 7)) >> 7;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is;
			dst += os;
		}
	}

    return 0;
}

int pyc_to_nv12s(unsigned char *dst, const unsigned short *src,
                 unsigned int h, unsigned int w,
                 unsigned int is, unsigned int os)
{
    unsigned int i,j,k,l = is * 2;
    int y,c;
    const PIXEL_YC *src_p = (const PIXEL_YC *)src;

    w = (w + 1)/2;
    is = (is - w * 2);
    os = (os - w * 2);

    for(i=0; i<h; ++i){
        for(j=0; j<w; ++j){
            y = src_p->y >> 4;
            *dst++ = CLIP(y);
            src_p++;
            y = src_p->y >> 4;
            *dst++ = CLIP(y);
            src_p++;
        }
        src_p += is;
        dst += os;
    }

	src_p = (const PIXEL_YC *)src;
	h /= 2;
	k = h & 1;
	h -= k;
    for(i=0; i<h; ++i){
		if(i&1){ //bottom
			for(j=0; j<w; ++j){
				//c = (src_p->cb + 3 * (src_p + l)->cb)/4;
				c = (src_p + l)->cb;
				c = (c + 2048) >> 4;
				*dst++ = CLIP(c);
				//c = (src_p->cr + 3 * (src_p + l)->cr)/4;
				c = (src_p + l)->cr;
				c = (c + 2048) >> 4;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is + l;
			dst += os;
		}else{ //top
			for(j=0; j<w; ++j){
				//c = (3 * src_p->cb + (src_p + l)->cb)/4;
				c = src_p->cb;
				c = (c + 2048) >> 4;
				*dst++ = CLIP(c);
				//c = (3 * src_p->cr + (src_p + l)->cr)/4;
				c = src_p->cr;
				c = (c + 2048) >> 4;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is;
			dst += os;
		}
    }
	if(k){
		for(i=0; i<2; ++i){
			for(j=0; j<w; ++j){
				c = (src_p->cb + 2048) >> 4;
				*dst++ = CLIP(c);
				c = (src_p->cr + 2048) >> 4;
				*dst++ = CLIP(c);
				src_p++;
				src_p++;
			}
			src_p += is;
			dst += os;
		}
	}

    return 0;
}
